/* -*- mode: c; c-basic-offset: 4; -*- */
#ifndef CHANNEL_H
#define CHANNEL_H

#include <sys/types.h>

/** \ingroup daq2_api
  * \{
  */

/**  Define channel types. The channel types are used to distinguish 
  *  the requested source of the data. 
  *  \remarks
  *  This expands on and replaces the channel group code in the NDS1 API
  *  which seems to have a very few values (0=normal fast channel, 1000=dmt 
  *  trend channel, 1001=obsolete dmt channel).
  *  \brief Channel type code enumerator.
  */
typedef enum chantype {

    /**  Unknown or unspecified default type. */
    cUnknown,

    /**  Online channel */
    cOnline,

    /**  Archived raw data channel */
    cRaw,

    /** Processed/RDS data channel */
    cRDS,

    /**  Second trend data */
    cSTrend,

    /**  Minute trend data */
    cMTrend,

    /**  Minute trend data */
    cTestPoint

} chantype_t;

/** Allowed maximum length for DMT channels */
#define MAX_LONG_CHANNEL_NAME_LENGTH 255

/** Allowed maximum length for DAQ channels */
#define MAX_CHANNEL_NAME_LENGTH 64

/**  Data type enumerator.
  *  \remarks numbering must be contiguous 
  */
typedef enum {
    /** The data type is not defined. */
    _undefined = 0,

    /** 16-bit (short) integer data. */
    _16bit_integer = 1,

    /** 32-bit (int) integer data. */
    _32bit_integer = 2,

    /** 64-bit (long) integer data. */
    _64bit_integer = 3,

    /** 32-bit (float) floating point data. */
    _32bit_float = 4,

    /** 64-bit (double) floating point data. */
    _64bit_double = 5,

    /** Complex data from two 32-bit floats {re, im}. */
    _32bit_complex = 6
} daq_data_t;

/** should be equal to the first data type
  */
#define MIN_DATA_TYPE _16bit_integer

/** should be equal to the last data type
  */
#define MAX_DATA_TYPE _32bit_complex

/**  Convert a channel type string to a channel type code
 */
chantype_t cvt_str_chantype(const char* str);

/**  Convert a channel type code to a string.
 */
const char* cvt_chantype_str(chantype_t code);

/** Function to relate data types to size.
  */
size_t data_type_size(daq_data_t dtype);

/** Function to relate data types to maximum value.
  */
double data_type_max(daq_data_t dtype);

/** Function to return data type name.
  */
const char* data_type_name(daq_data_t dtype);

/** Function to return data type name.
  */
daq_data_t data_type_code(const char* name);

/**  Inline function to return word lengthin bytes. This number gives the
  *  granularity in byte for use when swapping data bytes. It differs from
  *  the size attribute for complex numbers.
  */
size_t data_type_word(daq_data_t dtype);

/** \}
  */

#endif
